#include <functional>
#include <future>
#include <memory>
#include <string>
#include <sstream>

#include <rclcpp/rclcpp.hpp>
#include <rclcpp_action/rclcpp_action.hpp>

#include <franka_cps_msgs/action/grasp_sequence.hpp>
#include <geometry_msgs/msg/pose.hpp>

#include <moveit/planning_scene/planning_scene.h>
#include <moveit/task_constructor/task.h>
#include <moveit/task_constructor/stages.h>

#include <moveit/task_constructor/solvers.h>


static const rclcpp::Logger LOGGER = rclcpp::get_logger("franka_grasp_interface");
namespace mtc = moveit::task_constructor;

class GraspInterfaceNode : public rclcpp::Node
{
    public:
        using GraspSequence = franka_cps_msgs::action::GraspSequence;
        using GoalHandleGraspSequence = rclcpp_action::ServerGoalHandle<GraspSequence>;

        GraspInterfaceNode()
        : Node("grasp_interface")
        {
            // Action server for getting new Grasp Sequences
            action_server_ = rclcpp_action::create_server<GraspSequence>(
                this,
                "franka_grasp_moveit/grasp",
                std::bind(&GraspInterfaceNode::handle_goal, this, std::placeholders::_1, std::placeholders::_2),
                std::bind(&GraspInterfaceNode::handle_cancel, this, std::placeholders::_1),
                std::bind(&GraspInterfaceNode::handle_accepted, this, std::placeholders::_1));

            // Setup for move groups
            static const std::string arm_group_name = "panda_arm";
            static const std::string hand_group_name = "hand";
            static const std::string hand_frame = "panda_link8";
            static const std::string eef_name = "hand";
             

        }

    private:
        rclcpp_action::Server<GraspSequence>::SharedPtr action_server_;
        static const std::string arm_group_name;
        static const std::string hand_group_name;
        static const std::string hand_frame;
        static const std::string eef_name;
        const std::array<double, 6> tcp_frame = {0, 0, 0.1034, 0, 0, 0}; 

        rclcpp_action::GoalResponse handle_goal(
            const rclcpp_action::GoalUUID & uuid,
            std::shared_ptr<const GraspSequence::Goal> /*goal*/)
        {
            RCLCPP_INFO(this->get_logger(), "Received goal request!");
            (void)uuid;
            return rclcpp_action::GoalResponse::ACCEPT_AND_EXECUTE;
        }

        rclcpp_action::CancelResponse handle_cancel(const std::shared_ptr<GoalHandleGraspSequence> goal_handle)
        {
            RCLCPP_INFO(this->get_logger(), "Received request to cancel goal");
            (void)goal_handle;
            return rclcpp_action::CancelResponse::ACCEPT;
        }   
        
        void handle_accepted(const std::shared_ptr<GoalHandleGraspSequence> goal_handle) 
        {
        // This needs to run in a separate thread to avoid blocking the executor
        std::thread([this, goal_handle]() {
            const auto goal = goal_handle->get_goal();
            auto result = std::make_shared<GraspSequence::Result>();

            // Generate and execute the task
            try {
                auto task = createTask(*goal);
                auto success = executeTask(task);
                
                if (success) {
                    result->result_msg = "Execution successful";
                    result->error_code = 0; // Success code
                    goal_handle->succeed(result);
                } else {
                    result->result_msg = "Execution failed";
                    result->error_code = 1; // Generic failure code
                    goal_handle->abort(result);
                }
            } catch (const std::exception& e) {
                RCLCPP_ERROR(this->get_logger(), "Execution exception: %s", e.what());
                result->result_msg = std::string("Execution exception: ") + e.what();
                result->error_code = 2; // Exception code
                goal_handle->abort(result);
            }
        }).detach();
        }
        
        // Generating a task
        auto createTask(const GraspSequence::Goal& goal) -> moveit::task_constructor::Task {
            moveit::task_constructor::Task task;
            task.stages()->setName("Grasp Task");
            task.loadRobotModel(this->shared_from_this());

            // set Planners
            auto sampling_planner = std::make_shared<mtc::solvers::PipelinePlanner>(this->shared_from_this());
            auto interpolation_planner = std::make_shared<mtc::solvers::JointInterpolationPlanner>();
            auto cartesian_planner = std::make_shared<mtc::solvers::CartesianPath>();
            cartesian_planner->setMaxVelocityScalingFactor(1.0);
            cartesian_planner->setMaxAccelerationScalingFactor(1.0);
            cartesian_planner->setStepSize(.01);

            // Set task properties
            task.setProperty("group", arm_group_name);
            task.setProperty("eef", eef_name);
            task.setProperty("hand", hand_group_name);
            task.setProperty("hand_grasping_frame", hand_frame);
            task.setProperty("ik_frame", hand_frame);

            // Current State
            auto current_state = std::make_unique<moveit::task_constructor::stages::CurrentState>("current state");

            // Open Hand
            moveit::task_constructor::Stage* initial_state_ptr = nullptr;

            auto stage1 = std::make_unique<moveit::task_constructor::stages::MoveTo>("open hand", sampling_planner);
            stage1.setGroup(hand_group_name);
            stage1.setGoal("open");
            initial_state_ptr = stage1.get();
            task.add(std::move(stage1));



            // Grasping move
            auto grasp = std::make_unique<moveit::task_constructor::SerialContainer>("pick object");
		    task.properties().exposeTo(grasp.properties(), { "eef", "hand", "group", "ik_frame" });
		    grasp.properties().configureInitFrom(moveit::task_constructor::Stage::PARENT, { "eef", "hand", "group", "ik_frame" });

            // Move Hower
            auto stage2 = std::make_unique<moveit::task_constructor::stages::GenerateGraspPose>("generate hower pose");
			stage2.properties().configureInitFrom(moveit::task_constructor::Stage::PARENT);
			stage2.properties().set("marker_ns", "hower_pose");
			stage2.setPose(goal.hower)
			stage2.setMonitoredStage(initial_state_ptr);  // hook into successful initial-phase solutions

			// Compute IK Hower
			auto wrapper1 = std::make_unique<moveit::task_constructor::stages::ComputeIK>("hower pose IK", std::move(stage));
			wrapper1.setMaxIKSolutions(2);
			wrapper1.setMinSolutionDistance(1.0);
			wrapper1.setIKFrame(vectorToEigen(tcp_frame), hand_frame);
			wrapper1.properties().configureInitFrom(moveit::task_constructor::Stage::PARENT, { "eef", "group" });
			wrapper1.properties().configureInitFrom(moveit::task_constructor::Stage::INTERFACE, { "target_pose" });
			grasp.insert(std::move(wrapper1));

            // Move Grasp
            auto stage3 = std::make_unique<moveit::task_constructor::stages::GenerateGraspPose>("generate grasp pose");
			stage3.properties().configureInitFrom(moveit::task_constructor::Stage::PARENT);
			stage3.properties().set("marker_ns", "grasp_pose");
			stage3.setPose(goal.grasp)
			stage3.setMonitoredStage(initial_state_ptr);  // hook into successful initial-phase solutions

			// Compute IK Grasp
			auto wrapper2 = std::make_unique<moveit::task_constructor::stages::ComputeIK>("grasp pose IK", std::move(stage));
			wrapper2.setMaxIKSolutions(2);
			wrapper2.setMinSolutionDistance(1.0);
			wrapper2.setIKFrame(vectorToEigen(tcp_frame), hand_frame);
			wrapper2.properties().configureInitFrom(moveit::task_constructor::Stage::PARENT, { "eef", "group" });
			wrapper2.properties().configureInitFrom(moveit::task_constructor::Stage::INTERFACE, { "target_pose" });
			grasp.insert(std::move(wrapper2));

            // Allow Collision
            auto stage4 = std::make_unique<moveit::task_constructor::stages::ModifyPlanningScene>("allow collision (hand,object)");
			stage4.allowCollisions(
			    "object",
			    task.getRobotModel().getJointModelGroup(hand_group_name).getLinkModelNamesWithCollisionGeometry(),
			    true);
			grasp.insert(std::move(stage4));

            // Close Hand
            auto stage5 = std::make_unique<moveit::task_constructor::stages::MoveTo>("close hand", sampling_planner);
			stage5.setGroup(hand_group_name);
			stage5.setGoal("close");
			grasp.insert(std::move(stage5));

            // Move Back to Hower
            auto stage6 = std::make_unique<moveit::task_constructor::stages::GenerateGraspPose>("generate hower pose");
			stage6.properties().configureInitFrom(Stage::PARENT);
			stage6.properties().set("marker_ns", "hower_pose");
			stage6.setPose(goal.hower)
			stage6.setMonitoredStage(initial_state_ptr);  // hook into successful initial-phase solutions

			// Compute IK Hower
			grasp.insert(std::move(wrapper2));

            // Transition to Drop Phase
            task.add(std::move(grasp));


            // Move Drop
            auto stage7 = std::make_unique<moveit::task_constructor::stages::GenerateGraspPose>("generate hower pose");
			stage7.properties().configureInitFrom(moveit::task_constructor::Stage::PARENT);
			stage7.properties().set("marker_ns", "hower_pose");
			stage7.setPose(goal.drop)
			stage7.setMonitoredStage(initial_state_ptr);  // hook into successful initial-phase solutions

			// Compute IK Drop
			auto wrapper3 = std::make_unique<moveit::task_constructor::stages::ComputeIK>("drop pose IK", std::move(stage));
			wrapper3.setMaxIKSolutions(2);
			wrapper3.setMinSolutionDistance(1.0);
			wrapper3.setIKFrame(vectorToEigen(tcp_frame), hand_frame);
			wrapper3.properties().configureInitFrom(moveit::task_constructor::Stage::PARENT, { "eef", "group" });
			wrapper3.properties().configureInitFrom(moveit::task_constructor::Stage::INTERFACE, { "target_pose" });
			task.add(std::move(wrapper3));

            // Open Hand
            auto stage8 = std::make_unique<moveit::task_constructor::stages::MoveTo>("open hand", sampling_planner);
			stage8.setGroup(hand_group_name);
			stage8.setGoal("open");
			task.add(std::move(stage8));

            // Move Idle
            auto stage9 = std::make_unique<moveit::task_constructor::stages::GenerateGraspPose>("generate idle pose");
			stage9.properties().configureInitFrom(moveit::task_constructor::Stage::PARENT);
			stage9.properties().set("marker_ns", "hower_pose");
			stage9.setPose(goal.idle)
			stage9.setMonitoredStage(initial_state_ptr);  // hook into successful initial-phase solutions

			// Compute IK Idle
			auto wrapper4 = std::make_unique<moveit::task_constructor::stages::ComputeIK>("drop pose IK", std::move(stage));
			wrapper4.setMaxIKSolutions(2);
			wrapper4.setMinSolutionDistance(1.0);
			wrapper4.setIKFrame(vectorToEigen(tcp_frame), hand_frame);
			wrapper4.properties().configureInitFrom(moveit::task_constructor::Stage::PARENT, { "eef", "group" });
			wrapper4.properties().configureInitFrom(moveit::task_constructor::Stage::INTERFACE, { "target_pose" });
			task.add(std::move(wrapper4));

            return task;
        }

        bool executeTask(moveit::task_constructor::Task::Task task) {
            RCLCPP_INFO(LOGGER, "Executing solution trajectory");
            moveit_msgs::msg::MoveItErrorCodes execute_result;

            execute_result = task.execute(*task.solutions().front());

            if (execute_result.val != moveit_msgs::msg::MoveItErrorCodes::SUCCESS) {
                RCLCPP_ERROR_STREAM(LOGGER, "Task execution failed and returned: " << execute_result.val);
                return false;
	        }

            return true;
        }
};


int main(int argc, char** argv)
{
    rclcpp::init(argc, argv);

    rclcpp::shutdown();
    return 0;
}